@@ -16,19 +16,22 @@
 #define INLINE inline
 #endif
 
+#include "PosixSource.h"
 #include "Rts.h"
+
 #include "RtsUtils.h"
 #include "RetainerProfile.h"
 #include "RetainerSet.h"
 #include "Schedule.h"
 #include "Printer.h"
-#include "RtsFlags.h"
 #include "Weak.h"
-#include "Sanity.h"
+#include "sm/Sanity.h"
 #include "Profiling.h"
 #include "Stats.h"
 #include "ProfHeap.h"
 #include "Apply.h"
+#include "Stable.h" /* markStableTables */
+#include "sm/Storage.h" // for END_OF_STATIC_LIST
 
 /*
   Note: what to change in order to plug-in a new retainer profiling scheme?
@@ -269,11 +272,11 @@ isEmptyRetainerStack( void )
  * Returns size of stack
  * -------------------------------------------------------------------------- */
 #ifdef DEBUG
-lnat
+W_
 retainerStackBlocks( void )
 {
     bdescr* bd;
-    lnat res = 0;
+    W_ res = 0;
 
     for (bd = firstStack; bd != NULL; bd = bd->link) 
       res += bd->blocks;
@@ -364,7 +367,7 @@ find_srt( stackPos *info )
 	bitmap = info->next.srt.srt_bitmap;
 	while (bitmap != 0) {
 	    if ((bitmap & 1) != 0) {
-#if defined(__PIC__) && defined(mingw32_TARGET_OS)
+#if defined(COMPILING_WINDOWS_DLL)
 		if ((unsigned long)(*(info->next.srt.srt)) & 0x1)
 		    c = (* (StgClosure **)((unsigned long)*(info->next.srt.srt)) & ~0x1);
 		else
@@ -451,10 +454,6 @@ push( StgClosure *c, retainer c_child_r, StgClosure **first_child )
 	// no child, no SRT
     case CONSTR_0_1:
     case CONSTR_0_2:
-    case CAF_BLACKHOLE:
-    case BLACKHOLE:
-    case SE_BLACKHOLE:
-    case SE_CAF_BLACKHOLE:
     case ARR_WORDS:
 	*first_child = NULL;
 	return;
@@ -468,8 +467,7 @@ push( StgClosure *c, retainer c_child_r, StgClosure **first_child )
 	*first_child = ((StgSelector *)c)->selectee;
 	return;
     case IND_PERM:
-    case IND_OLDGEN_PERM:
-    case IND_OLDGEN:
+    case BLACKHOLE:
 	*first_child = ((StgInd *)c)->indirectee;
 	return;
     case CONSTR_1_0:
@@ -508,8 +506,10 @@ push( StgClosure *c, retainer c_child_r, StgClosure **first_child )
 	break;
 
 	// layout.payload.ptrs, no SRT
+    case TVAR:
     case CONSTR:
-    case STABLE_NAME:
+    case PRIM:
+    case MUT_PRIM:
     case BCO:
     case CONSTR_STATIC:
 	init_ptrs(&se.info, get_itbl(c)->layout.payload.ptrs,
@@ -589,16 +589,6 @@ push( StgClosure *c, retainer c_child_r, StgClosure **first_child )
 	    return;     // no child
 	break;
 	
-    case TVAR_WATCH_QUEUE:
-	*first_child = (StgClosure *)((StgTVarWatchQueue *)c)->closure;
-	se.info.next.step = 2;            // 2 = second
-	break;
-    case TVAR:
-	*first_child = (StgClosure *)((StgTVar *)c)->current_value;
-	break;
-    case TREC_HEADER:
-	*first_child = (StgClosure *)((StgTRecHeader *)c)->enclosing_trec;
-	break;
     case TREC_CHUNK:
 	*first_child = (StgClosure *)((StgTRecChunk *)c)->prev_chunk;
 	se.info.next.step = 0;  // entry no.
@@ -609,23 +599,19 @@ push( StgClosure *c, retainer c_child_r, StgClosure **first_child )
     case AP:
     case AP_STACK:
     case TSO:
+    case STACK:
     case IND_STATIC:
     case CONSTR_NOCAF_STATIC:
 	// stack objects
     case UPDATE_FRAME:
     case CATCH_FRAME:
+    case UNDERFLOW_FRAME:
     case STOP_FRAME:
-    case RET_DYN:
     case RET_BCO:
     case RET_SMALL:
     case RET_BIG:
 	// invalid objects
     case IND:
-    case BLOCKED_FETCH:
-    case FETCH_ME:
-    case FETCH_ME_BQ:
-    case RBH:
-    case REMOTE_REF:
     case INVALID_OBJECT:
     default:
 	barf("Invalid object *c in push()");
@@ -832,33 +818,6 @@ pop( StgClosure **c, StgClosure **cp, retainer *r )
 	    *r = se->c_child_r;
 	    return;
 
-	case TVAR_WATCH_QUEUE:
-	    if (se->info.next.step == 2) {
-		*c = (StgClosure *)((StgTVarWatchQueue *)se->c)->next_queue_entry;
-		se->info.next.step++;             // move to the next step
-		// no popOff
-	    } else {
-		*c = (StgClosure *)((StgTVarWatchQueue *)se->c)->prev_queue_entry;
-		popOff();
-	    }
-	    *cp = se->c;
-	    *r = se->c_child_r;
-	    return;
-
-	case TVAR:
-	    *c = (StgClosure *)((StgTVar *)se->c)->first_watch_queue_entry;
-	    *cp = se->c;
-	    *r = se->c_child_r;
-	    popOff();
-	    return;
-
-	case TREC_HEADER:
-	    *c = (StgClosure *)((StgTRecHeader *)se->c)->current_chunk;
-	    *cp = se->c;
-	    *r = se->c_child_r;
-	    popOff();
-	    return;
-
 	case TREC_CHUNK: {
 	    // These are pretty complicated: we have N entries, each
 	    // of which contains 3 fields that we want to follow.  So
@@ -887,8 +846,10 @@ pop( StgClosure **c, StgClosure **cp, retainer *r )
 	    return;
 	}
 
-	case CONSTR:
-	case STABLE_NAME:
+        case TVAR:
+        case CONSTR:
+	case PRIM:
+	case MUT_PRIM:
 	case BCO:
 	case CONSTR_STATIC:
 	    // StgMutArrPtr.ptrs, no SRT
@@ -956,41 +917,31 @@ pop( StgClosure **c, StgClosure **cp, retainer *r )
 	    // no child (fixed), no SRT
 	case CONSTR_0_1:
 	case CONSTR_0_2:
-	case CAF_BLACKHOLE:
-	case BLACKHOLE:
-	case SE_BLACKHOLE:
-	case SE_CAF_BLACKHOLE:
 	case ARR_WORDS:
 	    // one child (fixed), no SRT
 	case MUT_VAR_CLEAN:
 	case MUT_VAR_DIRTY:
 	case THUNK_SELECTOR:
 	case IND_PERM:
-	case IND_OLDGEN_PERM:
-	case IND_OLDGEN:
 	case CONSTR_1_1:
 	    // cannot appear
 	case PAP:
 	case AP:
 	case AP_STACK:
 	case TSO:
-	case IND_STATIC:
+        case STACK:
+        case IND_STATIC:
 	case CONSTR_NOCAF_STATIC:
 	    // stack objects
-	case RET_DYN:
-	case UPDATE_FRAME:
+        case UPDATE_FRAME:
 	case CATCH_FRAME:
-	case STOP_FRAME:
+        case UNDERFLOW_FRAME:
+        case STOP_FRAME:
 	case RET_BCO:
 	case RET_SMALL:
 	case RET_BIG:
 	    // invalid objects
 	case IND:
-	case BLOCKED_FETCH:
-	case FETCH_ME:
-	case FETCH_ME_BQ:
-	case RBH:
-	case REMOTE_REF:
 	case INVALID_OBJECT:
 	default:
 	    barf("Invalid object *c in pop()");
@@ -1055,16 +1006,17 @@ isRetainer( StgClosure *c )
 	//
 	// TSOs MUST be retainers: they constitute the set of roots.
     case TSO:
+    case STACK:
 
 	// mutable objects
+    case MUT_PRIM:
     case MVAR_CLEAN:
     case MVAR_DIRTY:
+    case TVAR:
     case MUT_VAR_CLEAN:
     case MUT_VAR_DIRTY:
     case MUT_ARR_PTRS_CLEAN:
     case MUT_ARR_PTRS_DIRTY:
-    case MUT_ARR_PTRS_FROZEN:
-    case MUT_ARR_PTRS_FROZEN0:
 
 	// thunks are retainers.
     case THUNK:
@@ -1083,10 +1035,6 @@ isRetainer( StgClosure *c )
 	// WEAK objects are roots; there is separate code in which traversing
 	// begins from WEAK objects.
     case WEAK:
-
-	// Since the other mutvar-type things are retainers, seems
-	// like the right thing to do:
-    case TVAR:
 	return rtsTrue;
 
 	//
@@ -1109,33 +1057,30 @@ isRetainer( StgClosure *c )
     case FUN_0_2:
 	// partial applications
     case PAP:
-	// blackholes
-    case CAF_BLACKHOLE:
-    case BLACKHOLE:
-    case SE_BLACKHOLE:
-    case SE_CAF_BLACKHOLE:
 	// indirection
     case IND_PERM:
-    case IND_OLDGEN_PERM:
-    case IND_OLDGEN:
+    // IND_STATIC used to be an error, but at the moment it can happen
+    // as isAlive doesn't look through IND_STATIC as it ignores static
+    // closures. See trac #3956 for a program that hit this error.
+    case IND_STATIC:
+    case BLACKHOLE:
 	// static objects
     case CONSTR_STATIC:
     case FUN_STATIC:
 	// misc
-    case STABLE_NAME:
+    case PRIM:
     case BCO:
     case ARR_WORDS:
 	// STM
-    case TVAR_WATCH_QUEUE:
-    case TREC_HEADER:
     case TREC_CHUNK:
+        // immutable arrays
+    case MUT_ARR_PTRS_FROZEN:
+    case MUT_ARR_PTRS_FROZEN0:
 	return rtsFalse;
 
 	//
 	// Error case
 	//
-	// IND_STATIC cannot be *c, *cp, *r in the retainer profiling loop.
-    case IND_STATIC:
 	// CONSTR_NOCAF_STATIC
 	// cannot be *c, *cp, *r in the retainer profiling loop.
     case CONSTR_NOCAF_STATIC:
@@ -1143,18 +1088,13 @@ isRetainer( StgClosure *c )
 	// legal objects during retainer profiling.
     case UPDATE_FRAME:
     case CATCH_FRAME:
+    case UNDERFLOW_FRAME:
     case STOP_FRAME:
-    case RET_DYN:
     case RET_BCO:
     case RET_SMALL:
     case RET_BIG:
 	// other cases
     case IND:
-    case BLOCKED_FETCH:
-    case FETCH_ME:
-    case FETCH_ME_BQ:
-    case RBH:
-    case REMOTE_REF:
     case INVALID_OBJECT:
     default:
 	barf("Invalid object in isRetainer(): %d", get_itbl(c)->type);
@@ -1297,9 +1237,9 @@ retainSRT (StgClosure **srt, nat srt_bitmap, StgClosure *c, retainer c_child_r)
 
   while (bitmap != 0) {
       if ((bitmap & 1) != 0) {
-#ifdef ENABLE_WIN32_DLL_SUPPORT
+#if defined(COMPILING_WINDOWS_DLL)
 	  if ( (unsigned long)(*srt) & 0x1 ) {
-	      retainClosure(*stgCast(StgClosure**,(stgCast(unsigned long, *srt) & ~0x1)), 
+	      retainClosure(* (StgClosure**) ((unsigned long) (*srt) & ~0x1), 
 			    c, c_child_r);
 	  } else {
 	      retainClosure(*srt,c,c_child_r);
@@ -1325,8 +1265,8 @@ retainSRT (StgClosure **srt, nat srt_bitmap, StgClosure *c, retainer c_child_r)
  *    RSET(c) and RSET(c_child_r) are valid, i.e., their
  *    interpretation conforms to the current value of flip (even when they
  *    are interpreted to be NULL).
- *    If *c is TSO, its state is not any of ThreadRelocated, ThreadComplete,
- *    or ThreadKilled, which means that its stack is ready to process.
+ *    If *c is TSO, its state is not ThreadComplete,or ThreadKilled, 
+ *    which means that its stack is ready to process.
  *  Note:
  *    This code was almost plagiarzied from GC.c! For each pointer,
  *    retainClosure() is invoked instead of evacuate().
@@ -1338,7 +1278,7 @@ retainStack( StgClosure *c, retainer c_child_r,
     stackElement *oldStackBoundary;
     StgPtr p;
     StgRetInfoTable *info;
-    StgWord32 bitmap;
+    StgWord bitmap;
     nat size;
 
 #ifdef DEBUG_RETAINER
@@ -1359,11 +1299,8 @@ retainStack( StgClosure *c, retainer c_child_r,
     // debugBelch("retainStack() called: oldStackBoundary = 0x%x, currentStackBoundary = 0x%x\n", oldStackBoundary, currentStackBoundary);
 #endif
 
-    ASSERT(get_itbl(c)->type != TSO || 
-	   (((StgTSO *)c)->what_next != ThreadRelocated &&
-	    ((StgTSO *)c)->what_next != ThreadComplete &&
-	    ((StgTSO *)c)->what_next != ThreadKilled));
-    
+    ASSERT(get_itbl(c)->type == STACK);
+
     p = stackStart;
     while (p < stackEnd) {
 	info = get_ret_itbl((StgClosure *)p);
@@ -1375,7 +1312,8 @@ retainStack( StgClosure *c, retainer c_child_r,
 	    p += sizeofW(StgUpdateFrame);
 	    continue;
 
-	case STOP_FRAME:
+        case UNDERFLOW_FRAME:
+        case STOP_FRAME:
 	case CATCH_FRAME:
 	case CATCH_STM_FRAME:
 	case CATCH_RETRY_FRAME:
@@ -1413,29 +1351,7 @@ retainStack( StgClosure *c, retainer c_child_r,
 	    // and don't forget to follow the SRT 
 	    goto follow_srt;
 
-	    // Dynamic bitmap: the mask is stored on the stack 
-	case RET_DYN: {
-	    StgWord dyn;
-	    dyn = ((StgRetDyn *)p)->liveness;
-
-	    // traverse the bitmap first
-	    bitmap = RET_DYN_LIVENESS(dyn);
-	    p      = (P_)&((StgRetDyn *)p)->payload[0];
-	    size   = RET_DYN_BITMAP_SIZE;
-	    p = retain_small_bitmap(p, size, bitmap, c, c_child_r);
-	    
-	    // skip over the non-ptr words
-	    p += RET_DYN_NONPTRS(dyn) + RET_DYN_NONPTR_REGS_SIZE;
-	    
-	    // follow the ptr words
-	    for (size = RET_DYN_PTRS(dyn); size > 0; size--) {
-		retainClosure((StgClosure *)*p, c, c_child_r);
-		p++;
-	    }
-	    continue;
-	}
-
-	case RET_FUN: {
+        case RET_FUN: {
 	    StgRetFun *ret_fun = (StgRetFun *)p;
 	    StgFunInfoTable *fun_info;
 	    
@@ -1628,14 +1544,7 @@ inner_loop:
 #endif
 	    goto loop;
 	}
-	if (((StgTSO *)c)->what_next == ThreadRelocated) {
-#ifdef DEBUG_RETAINER
-	    debugBelch("ThreadRelocated encountered in retainClosure()\n");
-#endif
-	    c = (StgClosure *)((StgTSO *)c)->_link;
-	    goto inner_loop;
-	}
-	break;
+        break;
 
     case IND_STATIC:
 	// We just skip IND_STATIC, so its retainer set is never computed.
@@ -1749,12 +1658,30 @@ inner_loop:
     // than attempting to save the current position, because doing so
     // would be hard.
     switch (typeOfc) {
-    case TSO:
+    case STACK:
 	retainStack(c, c_child_r,
-		    ((StgTSO *)c)->sp,
-		    ((StgTSO *)c)->stack + ((StgTSO *)c)->stack_size);
+                    ((StgStack *)c)->sp,
+                    ((StgStack *)c)->stack + ((StgStack *)c)->stack_size);
 	goto loop;
 
+    case TSO:
+    {
+        StgTSO *tso = (StgTSO *)c;
+
+        retainClosure(tso->stackobj,           c, c_child_r);
+        retainClosure(tso->blocked_exceptions, c, c_child_r);
+        retainClosure(tso->bq,                 c, c_child_r);
+        retainClosure(tso->trec,               c, c_child_r);
+        if (   tso->why_blocked == BlockedOnMVar
+               || tso->why_blocked == BlockedOnMVarRead
+               || tso->why_blocked == BlockedOnBlackHole
+               || tso->why_blocked == BlockedOnMsgThrowTo
+            ) {
+            retainClosure(tso->block_info.closure, c, c_child_r);
+        }
+        goto loop;
+    }
+
     case PAP:
     {
 	StgPAP *pap = (StgPAP *)c;
@@ -1808,6 +1735,7 @@ retainRoot(void *user STG_UNUSED, StgClosure **tl)
     currentStackBoundary = stackTop;
 
     c = UNTAG_CLOSURE(*tl);
+    maybeInitRetainerSet(c);
     if (c != &stg_END_TSO_QUEUE_closure && isRetainer(c)) {
 	retainClosure(c, c, getRetainerFrom(c));
     } else {
@@ -1827,7 +1755,7 @@ computeRetainerSet( void )
 {
     StgWeak *weak;
     RetainerSet *rtl;
-    nat g;
+    nat g, n;
     StgPtr ml;
     bdescr *bd;
 #ifdef DEBUG_RETAINER
@@ -1841,12 +1769,15 @@ computeRetainerSet( void )
     //
     // The following code assumes that WEAK objects are considered to be roots
     // for retainer profilng.
-    for (weak = weak_ptr_list; weak != NULL; weak = weak->link)
-	// retainRoot((StgClosure *)weak);
-	retainRoot(NULL, (StgClosure **)&weak);
+    for (g = 0; g < RtsFlags.GcFlags.generations; g++) {
+        for (weak = generations[g].weak_ptr_list; weak != NULL; weak = weak->link) {
+            // retainRoot((StgClosure *)weak);
+            retainRoot(NULL, (StgClosure **)&weak);
+        }
+    }
 
     // Consider roots from the stable ptr table.
-    markStablePtrTable(retainRoot, NULL);
+    markStableTables(retainRoot, NULL);
 
     // The following code resets the rs field of each unvisited mutable
     // object (computing sumOfNewCostExtra and updating costArray[] when
@@ -1858,7 +1789,8 @@ computeRetainerSet( void )
 	// Traversing through mut_list is necessary
 	// because we can find MUT_VAR objects which have not been
 	// visited during retainer profiling.
-	for (bd = generations[g].mut_list; bd != NULL; bd = bd->link) {
+        for (n = 0; n < n_capabilities; n++) {
+          for (bd = capabilities[n]->mut_lists[g]; bd != NULL; bd = bd->link) {
 	    for (ml = bd->start; ml < bd->free; ml++) {
 
 		maybeInitRetainerSet((StgClosure *)*ml);
@@ -1889,7 +1821,8 @@ computeRetainerSet( void )
 		}
 #endif
 	    }
-	}
+          }
+        }
     }
 }
 
@@ -1908,6 +1841,15 @@ computeRetainerSet( void )
  *    However, this is not necessary because any static indirection objects
  *    are just traversed through to reach dynamic objects. In other words,
  *    they are not taken into consideration in computing retainer sets.
+ *
+ * SDM (20/7/2011): I don't think this is doing anything sensible,
+ * because it happens before retainerProfile() and at the beginning of
+ * retainerProfil() we change the sense of 'flip'.  So all of the
+ * calls to maybeInitRetainerSet() here are initialising retainer sets
+ * with the wrong flip.  Also, I don't see why this is necessary.  I
+ * added a maybeInitRetainerSet() call to retainRoot(), and that seems
+ * to have fixed the assertion failure in retainerSetOf() I was
+ * encountering.
  * -------------------------------------------------------------------------- */
 void
 resetStaticObjectForRetainerProfiling( StgClosure *static_objects )
