@@ -36,6 +36,14 @@
 # include <errno.h>
 #endif
 
+#ifdef HAVE_EVENTFD_H
+# include <sys/eventfd.h>
+#endif
+
+#ifdef HAVE_TERMIOS_H
+#include <termios.h>
+#endif
+
 #include <stdlib.h>
 #include <string.h>
 
@@ -57,6 +65,39 @@ static StgInt nHandlers = 0;    /* Size of handlers array */
 
 static nat n_haskell_handlers = 0;
 
+static sigset_t userSignals;
+static sigset_t savedSignals;
+
+#ifdef THREADED_RTS
+static Mutex sig_mutex; // protects signal_handlers, nHandlers
+#endif
+
+/* -----------------------------------------------------------------------------
+ * Initialisation / deinitialisation
+ * -------------------------------------------------------------------------- */
+
+void
+initUserSignals(void)
+{
+    sigemptyset(&userSignals);
+#ifdef THREADED_RTS
+    initMutex(&sig_mutex);
+#endif
+}
+
+void
+freeSignalHandlers(void) {
+    if (signal_handlers != NULL) {
+        stgFree(signal_handlers);
+        signal_handlers = NULL;
+        nHandlers = 0;
+        n_haskell_handlers = 0;
+    }
+#ifdef THREADED_RTS
+    closeMutex(&sig_mutex);
+#endif
+}
+
 /* -----------------------------------------------------------------------------
  * Allocate/resize the table of signal handlers.
  * -------------------------------------------------------------------------- */
@@ -82,18 +123,27 @@ more_handlers(int sig)
 }
 
 // Here's the pipe into which we will send our signals
-static int io_manager_pipe = -1;
+static int io_manager_wakeup_fd = -1;
+static int io_manager_control_fd = -1;
 
 #define IO_MANAGER_WAKEUP 0xff
 #define IO_MANAGER_DIE    0xfe
 #define IO_MANAGER_SYNC   0xfd
 
 void
-setIOManagerPipe (int fd)
+setIOManagerWakeupFd (int fd)
 {
     // only called when THREADED_RTS, but unconditionally
-    // compiled here because GHC.Conc depends on it.
-    io_manager_pipe = fd;
+    // compiled here because GHC.Event.Control depends on it.
+    io_manager_wakeup_fd = fd;
+}
+
+void
+setIOManagerControlFd (int fd)
+{
+    // only called when THREADED_RTS, but unconditionally
+    // compiled here because GHC.Event.Control depends on it.
+    io_manager_control_fd = fd;
 }
 
 void
@@ -101,44 +151,37 @@ ioManagerWakeup (void)
 {
     int r;
     // Wake up the IO Manager thread by sending a byte down its pipe
-    if (io_manager_pipe >= 0) {
+    if (io_manager_wakeup_fd >= 0) {
+#if defined(HAVE_EVENTFD)
+	StgWord64 n = (StgWord64)IO_MANAGER_WAKEUP;
+	r = write(io_manager_wakeup_fd, (char *) &n, 8);
+#else
 	StgWord8 byte = (StgWord8)IO_MANAGER_WAKEUP;
-	r = write(io_manager_pipe, &byte, 1);
+	r = write(io_manager_wakeup_fd, &byte, 1);
+#endif
         if (r == -1) { sysErrorBelch("ioManagerWakeup: write"); }
     }
 }
 
-void
-ioManagerSync (void)
-{
-    int r;
-    // Wake up the IO Manager thread by sending a byte down its pipe
-    if (io_manager_pipe >= 0) {
-	StgWord8 byte = (StgWord8)IO_MANAGER_SYNC;
-	r = write(io_manager_pipe, &byte, 1);
-        if (r == -1) { sysErrorBelch("ioManagerSync: write"); }
-    }
-}
-
 #if defined(THREADED_RTS)
 void
 ioManagerDie (void)
 {
     int r;
     // Ask the IO Manager thread to exit
-    if (io_manager_pipe >= 0) {
+    if (io_manager_control_fd >= 0) {
 	StgWord8 byte = (StgWord8)IO_MANAGER_DIE;
-	r = write(io_manager_pipe, &byte, 1);
+	r = write(io_manager_control_fd, &byte, 1);
         if (r == -1) { sysErrorBelch("ioManagerDie: write"); }
-        close(io_manager_pipe);
-        io_manager_pipe = -1;
+        io_manager_control_fd = -1;
+        io_manager_wakeup_fd = -1;
     }
 }
 
-Capability *
-ioManagerStartCap (Capability *cap)
+void
+ioManagerStartCap (Capability **cap)
 {
-    return rts_evalIO(cap,&base_GHCziConc_ensureIOManagerIsRunning_closure,NULL);
+    rts_evalIO(cap,&base_GHCziConcziIO_ensureIOManagerIsRunning_closure,NULL);
 }
 
 void
@@ -146,9 +189,9 @@ ioManagerStart (void)
 {
     // Make sure the IO manager thread is running
     Capability *cap;
-    if (io_manager_pipe < 0) {
+    if (io_manager_control_fd < 0 || io_manager_wakeup_fd < 0) {
 	cap = rts_lock();
-	cap = ioManagerStartCap(cap);
+        ioManagerStartCap(&cap);
 	rts_unlock(cap);
     }
 }
@@ -177,7 +220,7 @@ generic_handler(int sig USED_IF_THREADS,
 {
 #if defined(THREADED_RTS)
 
-    if (io_manager_pipe != -1)
+    if (io_manager_control_fd != -1)
     {
         StgWord8 buf[sizeof(siginfo_t) + 1];
         int r;
@@ -191,7 +234,7 @@ generic_handler(int sig USED_IF_THREADS,
 	    memcpy(buf+1, info, sizeof(siginfo_t));
 	}
 
-	r = write(io_manager_pipe, buf, sizeof(siginfo_t)+1);
+	r = write(io_manager_control_fd, buf, sizeof(siginfo_t)+1);
         if (r == -1 && errno == EAGAIN)
         {
             errorBelch("lost signal due to full pipe: %d\n", sig);
@@ -241,7 +284,7 @@ generic_handler(int sig USED_IF_THREADS,
 	stg_exit(EXIT_FAILURE);
     }
     
-    contextSwitchCapability(&MainCapability);
+    interruptCapability(&MainCapability);
 
 #endif /* THREADED_RTS */
 }
@@ -250,19 +293,6 @@ generic_handler(int sig USED_IF_THREADS,
  * Blocking/Unblocking of the user signals
  * -------------------------------------------------------------------------- */
 
-static sigset_t userSignals;
-static sigset_t savedSignals;
-
-void
-initUserSignals(void)
-{
-    sigemptyset(&userSignals);
-#ifndef THREADED_RTS
-    getStablePtr((StgPtr)&base_GHCziConc_runHandlers_closure); 
-    // needed to keep runHandler alive
-#endif
-}
-
 void
 blockUserSignals(void)
 {
@@ -306,11 +336,14 @@ stg_sig_install(int sig, int spi, void *mask)
     struct sigaction action;
     StgInt previous_spi;
 
+    ACQUIRE_LOCK(&sig_mutex);
+
     // Block the signal until we figure out what to do
     // Count on this to fail if the signal number is invalid
     if (sig < 0 || sigemptyset(&signals) ||
 	sigaddset(&signals, sig) || sigprocmask(SIG_BLOCK, &signals, &osignals)) {
-	return STG_SIG_ERR;
+        RELEASE_LOCK(&sig_mutex);
+        return STG_SIG_ERR;
     }
     
     more_handlers(sig);
@@ -350,7 +383,8 @@ stg_sig_install(int sig, int spi, void *mask)
     if (sigaction(sig, &action, NULL))
     {
         errorBelch("sigaction");
-	return STG_SIG_ERR;
+        RELEASE_LOCK(&sig_mutex);
+        return STG_SIG_ERR;
     }
 
     signal_handlers[sig] = spi;
@@ -375,9 +409,11 @@ stg_sig_install(int sig, int spi, void *mask)
     if (sigprocmask(SIG_SETMASK, &osignals, NULL))
     {
         errorBelch("sigprocmask");
-	return STG_SIG_ERR;
+        RELEASE_LOCK(&sig_mutex);
+        return STG_SIG_ERR;
     }
 
+    RELEASE_LOCK(&sig_mutex);
     return previous_spi;
 }
 
@@ -412,7 +448,7 @@ startSignalHandlers(Capability *cap)
 		       RtsFlags.GcFlags.initialStkSize, 
                        rts_apply(cap,
                                  rts_apply(cap,
-                                           &base_GHCziConc_runHandlers_closure,
+                                           &base_GHCziConcziSignal_runHandlers_closure,
                                            rts_mkPtr(cap, info)),
                                  rts_mkInt(cap, info->si_signo))));
   }
@@ -463,6 +499,86 @@ shutdown_handler(int sig STG_UNUSED)
     }
 }
 
+/* -----------------------------------------------------------------------------
+ * An empty signal handler, currently used for SIGPIPE
+ * -------------------------------------------------------------------------- */
+static void
+empty_handler (int sig STG_UNUSED)
+{
+    // nothing
+}
+
+/* -----------------------------------------------------------------------------
+   SIGTSTP handling
+
+   When a process is suspeended with ^Z and resumed again, the shell
+   makes no attempt to save and restore the terminal settings.  So on
+   resume, any terminal setting modificaions we made (e.g. turning off
+   ICANON due to hSetBuffering NoBuffering) may well be lost.  Hence,
+   we arrange to save and restore the terminal settings ourselves.
+
+   The trick we use is:
+     - catch SIGTSTP
+     - in the handler,  kill(getpid(),SIGSTOP)
+     - when this returns, restore the TTY settings
+   This means we don't have to catch SIGCONT too.
+
+   Note we don't re-throw SIGTSTP, we throw SIGSTOP instead.  This is
+   for a few reasons:
+
+      - re-throwing SIGTSTP would require temporarily restoring the
+        default sigaction.
+
+      - it doesn't work on certain buggy pthread implementations
+        (e.g. OpenBSD).
+
+      - throwing SIGTSTP seems slightly dodgy anyway.
+
+   -------------------------------------------------------------------------- */
+
+static void sigtstp_handler(int sig);
+static void set_sigtstp_action (rtsBool handle);
+
+static void
+sigtstp_handler (int sig STG_UNUSED)
+{
+    int fd;
+    struct termios ts[3];
+
+    // save the current TTY state for TTYs we modified
+    for (fd = 0; fd <= 2; fd++) {
+        if (__hscore_get_saved_termios(fd) != NULL) {
+            tcgetattr(fd,&ts[fd]);
+        }
+    }
+
+    // really stop the process now
+    kill(getpid(), SIGSTOP);
+
+    // on return, restore the TTY state
+    for (fd = 0; fd <= 2; fd++) {
+        if (__hscore_get_saved_termios(fd) != NULL) {
+            tcsetattr(0,TCSANOW,&ts[fd]);
+        }
+    }
+}
+
+static void
+set_sigtstp_action (rtsBool handle)
+{
+    struct sigaction sa;
+    if (handle) {
+        sa.sa_handler = sigtstp_handler;
+    } else {
+        sa.sa_handler = SIG_DFL;
+    }
+    sa.sa_flags = 0;
+    sigemptyset(&sa.sa_mask);
+    if (sigaction(SIGTSTP, &sa, NULL) != 0) {
+        sysErrorBelch("warning: failed to install SIGTSTP handler");
+    }
+}
+
 /* -----------------------------------------------------------------------------
  * Install default signal handlers.
  *
@@ -519,12 +635,16 @@ initDefaultHandlers(void)
 #endif
 
     // ignore SIGPIPE; see #1619
-    action.sa_handler = SIG_IGN;
+    // actually, we use an empty signal handler rather than SIG_IGN,
+    // so that SIGPIPE gets reset to its default behaviour on exec.
+    action.sa_handler = empty_handler;
     sigemptyset(&action.sa_mask);
     action.sa_flags = 0;
     if (sigaction(SIGPIPE, &action, &oact) != 0) {
 	sysErrorBelch("warning: failed to install SIGPIPE handler");
     }
+
+    set_sigtstp_action(rtsTrue);
 }
 
 void
@@ -544,13 +664,8 @@ resetDefaultHandlers(void)
     if (sigaction(SIGPIPE, &action, NULL) != 0) {
 	sysErrorBelch("warning: failed to uninstall SIGPIPE handler");
     }
-}
 
-void
-freeSignalHandlers(void) {
-    if (signal_handlers != NULL) {
-        stgFree(signal_handlers);
-    }
+    set_sigtstp_action(rtsFalse);
 }
 
 #endif /* RTS_USER_SIGNALS */
